; Parallel Port I/O Driver - PPIO
; Based on Waite book -
;  "Writing MS-DOS Device Drivers" by Robert Lai.
; John Wolff  26 Mar 93

; This version for LPT2 at 278h.

cseg           segment   para public 'code'
ppio           proc      far
               assume    cs:cseg, ds:cseg, es:cseg

; Request header data structures for implemented functions

rh             struc     ; fixed part of request header
rh_len         db ?      ; length of packet
rh_unit        db ?      ; unit code for block devices
rh_cmd         db ?      ; command code
rh_status      dw ?      ; status word returned
rh_res1        dd ?      ; 8 bytes reserved
rh_res2        dd ?
rh             ends

rh0            struc     ; header fields for request 0 - INIT
rh0_rh         db        size rh dup (?)
rh0_nunits     db ?      ; number of units for block devices
rh0_brk_ofs    dw ?      ; offset of break address
rh0_brk_seg    dw ?      ; segment ditto
rh0_bpb_tbo    dw ?      ; offset address of ptr to bpb array
rh0_bpb_tbs    dw ?      ; seg ditto
rh0_drv_ltr    db ?      ; first avail drive - dos 3, block devices
rh0            ends

rh4            struc     ; header fields for request 4 - INPUT
rh4_rh         db        size rh dup (?)
rh4_media      db ?      ; media descriptor byte
rh4_buf_ofs    dw ?      ; offset of transfer buffer address
rh4_buf_seg    dw ?      ; segment ditto
rh4_count      dw ?      ; block/character count
rh4_start      dw ?      ; transfer starting sector for block devices
rh4            ends

rh8            struc     ; header fields for request 8 & 9 - OUTPUT
rh8_rh         db        size rh dup (?)
rh8_media      db ?      ; media descriptor byte
rh8_buf_ofs    dw ?      ; offset of transfer buffer address
rh8_buf_seg    dw ?      ; segment ditto
rh8_count      dw ?      ; block/character count
rh8_start      dw ?      ; transfer starting sector for block devices
rh8            ends

begin:

; Device Header

next_dev       dd        -1             ; filled in by dos on installation
attributes     dw        8000h          ; basic character device
strategy       dw        dev_strategy   ; strategy routine address
interrupt      dw        dev_interrupt  ; interrupt routine address
dev_name       db        'LPT2    '     ; replaces standard LPT2

; Local Data Storage

rh_ofs         dw ?      ; offset part of request header address
rh_seg         dw ?      ; segment ditto

; Command jump table for all possible I/O requests

cmdtab         label     byte
               dw        init
               dw        media_check
               dw        get_bpb
               dw        ioctl_input
               dw        input
               dw        nd_input
               dw        input_status
               dw        input_flush
               dw        output
               dw        output_verify
               dw        output_status
               dw        output_flush
               dw        ioctl_out
               dw        open
               dw        close
               dw        removable
               dw        out_till_busy

; LPT2 Port Addresses

base           equ       278h           ; base address
porta          equ       base           ; data port
portb          equ       base+1         ; status port
portc          equ       base+2         ; control port

; Strategy Procedure

dev_strategy:  mov       cs:rh_seg, es       ; save seg address
               mov       cs:rh_ofs, bx       ; ditto offfset
               ret

; Interrupt Procedure

dev_interrupt:
               cld
               push      ds
               push      es
               push      ax
               push      bx
               push      cx
               push      dx
               push      di
               push      si
               mov       ax, cs:rh_seg       ; restore es ex strategy call
               mov       es, ax
               mov       bx, cs:rh_ofs

               ; select and jump to command routine
               mov       al, es:[bx].rh_cmd  ; get command code
               rol       al, 1               ; double it for word steps
               mov       ah, 0
               lea       di, cmdtab          ; start of command table
               add       di, ax
               jmp       word ptr [di]

; Command processing routines for every I/O request

init:          call      initial             ; actually at end of code
               lea       ax, initial         ; discard "initial" routine
               mov       es:[bx].rh0_brk_ofs, ax
               mov       es:[bx].rh0_brk_seg, cs
               jmp       done

media_check:
get_bpb:       jmp       done
ioctl_input:   jmp       error

input:
               mov       cx, es:[bx].rh4_count     ; load character count
               mov       di, es:[bx].rh4_buf_ofs   ; load buffer address
               mov       ax, es:[bx].rh4_buf_seg
               mov       es, ax
               mov       dx, portc                 ; load status port address
readl:         in        al, dx                    ; read from port
               mov       ah, 0h
               mov       es:[di], al               ; move to buffer
               inc       di
               loop      readl
               jmp       done

nd_input:
input_status:
input_flush:   jmp       done

output:
output_verify:
               mov       cx, es:[bx].rh8_count     ; load character count
               mov       di, es:[bx].rh8_buf_ofs   ; load buffer address
               mov       ax, es:[bx].rh8_buf_seg
               mov       es, ax
               mov       dx, porta                 ; load data port address
outl:          mov       al, byte ptr es:[di]      ; fetch char
               inc       di                        ; point to next
               out       dx, al                    ; write to port
               loop      outl
               jmp       done

output_status:
output_flush:  jmp       done
ioctl_out:     jmp       error
open:
close:         jmp       done
removable:
out_till_busy: jmp       error

; Error Exit

error:         or       es:[bx].rh_status, 8003h   ; set error bit and code
               jmp      done

; Common Exit

busy:          or        es:[bx].rh_status, 0200h  ; set busy bit

done:          or        es:[bx].rh_status, 0100h  ; set done bit
               pop       si
               pop       di
               pop       dx
               pop       cx
               pop       bx
               pop       ax
               pop       es
               pop       ds
               ret

; Initialisation Routine - discarded after loading.

initial        proc      near
               mov       ax, 0h    ; clear data port
               mov       dx, porta
               out       dx, al

               lea       dx, msg1  ; display message via dos func 9
               mov       ah, 09h
               int       21h
               ret

initial        endp

msg1           db  0dh,0ah,'JW''s PPIO Driver Loaded for LPT2 at 278h.',0dh,0ah,0dh,0ah,'$'

; The End

ppio           endp                ; end of ppio procedure
cseg           ends                ; end of cseg segment
               end       begin     ; end of program


